/*----------------------------------------------------------------------------------------

	Introduction to TigerSHARC Multiprocessing Systems Using VisualDSP++
	Maikel K.B., DSP-Europe 
	June 2002

	ID0.ASM

	This program demonstrates some of the Multiprocessing features handled by VisualDSP++ Tools.
	It shows the use of TigerSHARC in an MP System by performing the following tasks: 

		*	Setting up of control registers: SYSCON and SDRCON
	
		*	ID Checking routine to ensure that the code has been loaded into the right DSP

		*	Go into endless loop and wait for Vector Interrupt (VIPRT) to be generated by DSP ID1

		*	It serves the VIRPT Interrupt copying the contents of tx_ID0 to rx_ID1(internal 
			memory of DSP ID1 via the MMS)and to the shared external memory (shared_data buffer
			declared in SDRAM) using DMA transfers.
 	
		*	Finally, a broadcast write to all DSP's internal memory is performed. For this, 
			the bus lock feature and DMA transfers are used.

		*	After completion, it goes back into an endless loop.


------------------------------------------------------------------------------------------*/


#define TAPS 16
#define MMS_ID0 0x2000000		// ID0 Multiprocessor Memory Space (MMS) offset value	
#define MMS_ID1 0x2400000		// ID1 Multiprocessor Memory Space (MMS) offset value		
#define MMS_Broadcast 0x1C80000 // Broadcast Multiprocessor Memory Space (MMS) offset value 

.extern shared_data;			// Data array defined in external shared memory (SDRAM)
.extern rx_ID1;					// ID1 receive buffer (ID1.asm) for external port (DMA) data transfer from ID0

.global VIRPT_ISR_ID0;			// Address of Int. Serv. Rout. to be served by ID0 in response to 
								// VIRPT interrupt generated by ID1
.global rx_ID0;					// ID0 receive buffer for core (IALU) data transfer from ID1

/* DM data section in Memory Block 1 */
.section data1;
.align 4;
.var Broadcast_rx_ID0[TAPS];	// ID0 broadcast receive buffer for broadcast transfer from its own
.align 4; 
.var rx_ID0[TAPS];				// ID0 receive buffer for core (IALU) data transfer from ID1

/* DM data section in Memory Block 2 */
.section data2;
.align 4;
.var tx_ID0[TAPS]= "tx_ID0.dat";// ID0 transmit buffer for external port (DMA) data transfers to ID1 
								// and shared external memory (first half of buffer)

/* PM code in Memory Block 0 */
.section program;

start_ID0:
	
	/* Setting Up of Control Registers */

	j14 = j31 + 0x3a79e7;;			// Initial value after reset (0x279e7) modified 
    SYSCON = j14;;					// for 64-bit Bus width configuration

	j11 = j31 + 0x00005913;;		// Enable SDRAM: ENA=1, CAS=2 CL, pipedepth=0,			
	SDRCON = j11;;  				// page=512w, ref.rate=1200, trp=2, tras=5, init=1


	/* ID Checking Routine */

	xR0 = 0x3;;						// Routine to check that code is being loaded into the right DSP (ID0)
	xR1 = SYSTAT;;					// read SYSTAT value 
	xr2 = 0;;						// DSP = ID0
	xR1 = FEXT R1 by R0;;			// Extract processor ID value (bits 3-0) from SYSTAT
	xr1 = r2 - r1;;					// is this DSP with ID0?
	
	if NAEQ, jump incorrect_ID;;	// if not  correct, stop execution and enter endless loop


	/* DMA set up */
	
	XR3 = 0x00000000;;		
	XR1 = 0x00000000;;		
	XR2 = 0x00000000;;			
	XR3 = 0x00000000;;		

	XR1 = 0x00040004;;				// Count =4, modify = 4 for quad word accesses
	XR3 = 0x00000000;;				// Disable DMA
	DCD0 = XR3:0;;					// Set up DMA Channel 0 
	DCD1 = XR3:0;;					// Set up DMA Channel 1

	j0 = j31 + _dma_int;;			// Set up DMA0 & DMA1 interrupt vectors
	IVDMA0 = j0;;
	IVDMA1 = j0;;	

	/* Enable interrupts - DMA0, DMA1, VIRPT and GLOBAL int. */

	xR0 = IMASKH;;					// Read value of IMASKH
	xR1 = 0x10010000;;				// Global and VIRPT interrupts
	xR2 = R1 or R0;;				// OR with contents of IMASKH
	IMASKH = xR2;;					// Write back to IMASKH

 	xR3 = IMASKL;;					// Read value of IMASKL
	xR4 = 0x0000C000;;				// DMA0 and DMA1 interrupts
	xR5 = R3 or R4;;				// OR with contents of IMASKL
	IMASKL = xR5;;					// Write back to IMASKL

/* Terminate and wait */
wait0:   jump wait0;;				// will only exit the wait loop to serve the VIRPT interrupt generated by ID1


/* Vector Interrupt Service Routine , copy the contents of tx_ID0 to ext. mem. and rx_ID1 */

VIRPT_ISR_ID0:

	/* Set up and perform DMA transfers */

										// External port transmit - DMA0
	xr0 = tx_ID0;;						// xr0 = source index = int. mem. DSP_0
	xr1 = 0x00100004;;					// count = 0x10, modify = 4 (quadword)
	xr3 = 0x47000000;;					// int mem,prio=norm,2D=no,word=quad,int=yes,RQ=enbl,chain=no
	DCS0 = xr3:0;;						// Set up source for transfer

	yr0 = rx_ID1;;						// xr0 = destination index = ext. mem. = int. mem. DSP_1
	yr1 = 0x00100004;;					// count = 0x10, modify = 4 (quadword)
	yr3 = 0x87000000;;					// ext mem,prio=norm,2D=no,word=quad,int=yes,RQ=enbl,chain=no
	DCD0 = yr3:0;;						// Set up destination and start external port DMA0 transfer

	DCS1 = xr3:0;;						// External port transmit - DMA1. Set up source for transfer

	yr0 = shared_data;;					// xr0 = destination index = ext. mem.(SDRAM)
	DCD1 = yr3:0;;						// Set up destination and start external port DMA1 transfer


	/* Perform a BROADCAST write to ALL DSPs in the system (ID0 & ID1) */
								
	j11 = j31 + 1;;	// Lock bus before performing broadcast write in order to prevent external
	BUSLK = j11;; 	// devices from accessing data while being modified 

	BUS_MASTER:		if NBM, jump BUS_MASTER;;
	// if NOT current Bus Master wait until bus is acquired

	DCS0 = xr3:0;;				// External port transmit - DMA0. Set up source for transfer
	yr0 = MMS_Broadcast;;		// yr0 = destination index = ext. mem. = MMS Broadcast (ID0 & ID1)
	DCD0 = yr3:0;;				// Set up destination and start external port DMA0 transfer

	j11 = j31 + 0;;				// Clear BUSLK Bit 
	BUSLK = j11;; 				// Bus is again available for external port accesses					

	nop;;
	rti(ABS);nop;nop;nop;;

VIRPT_ISR_ID0.END:

_dma_int:

	nop;;nop;;nop;;rti(ABS);;	//	DMA0 & DMA1 Interrupt vector

incorrect_ID:

	nop;;nop;;nop;;jump incorrect_ID;;  	// ERROR - Code loaded into wrong DSP. Check DSP ID Number!!! 
